/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
                                  config.c

Loads and processes the soyMAIL configuration file.  This file is located using
the SOYMAIL_CONFIG logical name.  If the logical does not exist then soyMAIL
just continues on it's merry way.  Directives and comments must begin in the
first column.


CONFIGURATION DIRECTIVES
------------------------
[downtime]              disable the use of soyMAIL and give the specified
                        message to users as a simple announcement

[font-family-local]     additional font families presented in the selector
                        (intended for non-en language support)

[greeting]              "Welcome to soyMAIL - UNAUTHORISED USE IS PROHIBITED."

[help-about-site]       site-specific information presented on the Help
                        'About' page

[html-editor-load]      JavaScript to load and instantiate the HTML editor

[logout-realm]          for VMS Apache this enables the logout functionality
                        this string must be *exactly* the same as the AuthName
                        attribute of the authorization against soyMAIL
                        (a realm of '-' disables the logout button)

[message-list-footer]   site-specific information (HTML text) presented in a
                        separate panel below the message listing

[page-title]            superior line in main menu panel and titles all pages
[page-title1]           (synonym)

[page-title2]           when set the text in 'page-title1' is forced left
                        and the 'page-title2' if forced right on the page

[print-header]          text header for printed mail messages

[print-footer]          text footer for printed mail messages

[private-access]        allows mapping bewteen authenticated user
                        (CGI remote-user) and a VMS username

[public-access]         permits and maps request path strings to VMS Mail
                        e.g. /path/ /username/file/folder/

[public-options-default]  allows the soyMAIL administrator to set public
                          access 'personal option' settings
                          (required to escape each leading '[' with a '\')

[search-control]        'full'  (expensive and the default)
                        'header-only'  (very lightweight)
                        'keyword-only'  (no regular expressions allowed)
                        'none'  (very cheap :-)
                        'priority='  set priority of message text searches
                        'vms-select'  use VMS Mail select rather than search

[SMTP-default-host]     if an email address is provided without a domain part
                        an '@' and this string is appended to create an valid
                        (looking) RFC822 address - this disables the ability
                        to send VMS mail!

[SMTP-from-host]        used when constructing the 'user@from.host.name'

[SMTP-server-host]      host name/address of SMTP server soyMAIL connects to
                        for Internet mail transport (defaults to localhost) 

[soyMAIL-at-title]      site description provided in title bar of browser
                        window (defaults to "soyMAIL @ the.server.name")

[user-options-default]  allows the soyMAIL administrator to preset some
                        options (e.g. language) by providing options
                        configuration text aginst this directive
                        (required to escape each leading '[' with a '\')

[user-options-override]  allows the soyMAIL administrator to override user
                         selected options by providing options configuration
                         text against this directive
                         (required to escape each leading '[' with a '\')


CONDITIONAL CONFIGURATION
-------------------------
[if-CGI-<name>]         matches the value of the specifed CGI variable
[if-not-CGI-<name>]     not matches the value of the specifed CGI variable
[if-private]            if private mail access
[if-public]             if public mail access
[if-END]                following directives are unconditionally processed
[end]                   stop processing the configuration at this point
                       
If a match then uses the configuration directives up to the complementary
[if-END].  Matching is either done by keyword or regular expression (see brief
description in SEARCH.C module).  Regular expression are introduced by a
leading caret symbol ('^') which of course is not included in the expression.

Keyword example ...

  [if-CGI-http_host]  the.host.name
  [if-CGI-server_name]  the.server.name
  [if-not-CGI-remote_user]  R_J_ECUREUIL

Equivalent regular expression example ...

  [if-CGI-http_host]  ^*the\.host\.name*
  [if-CGI-server_name]  ^*the\.server\.name*
  [if-not-CGI-remote_user]  ^*R_J_ECUREUIL*

Regular expression equivalents of these common wildcards

  asterisk ('*')           .*
  question mark ('?')      .
  percentage ('%')         .

A non-existant CGI variable is treated as an empty string.  Non-empty and empty
strings may be respectively tested for with

  [if-CGI-variable_name]  ^.*
  [if-not-CGI-variable_name]  ^.*

Conditionals may be nested (to any depth) but every conditional *must* be
closed with an [if-END].  For example:

  [if-private]
  [page-title] The Page Title
  [if-CGI-server_name] www.site1.com
  [message-list-footer] For more information on Site 1 see ...
  [if-end]
  [if-CGI-server_name] www.site2.com
  [message-list-footer] For more information on Site 2 see ...
  [if-end]
  [if-end]
  #
  [if-public]
  [page-title] The Public Page
  [message-list-footer] Just an public example!
  [public-access] /public-2005/ /WEBMASTER/MAIL/PUB2005/
  [if-end]


PRIVATE-ACCESS
--------------
Each mapping has the following format

  <remote-user>/<auth-realm>/<vms-username>

where the <remote-user> is the username used for authentication purposes,
<auth-realm> is a WASDism and allows the script to know against which realm the
authentication occured, and <vms-username> is the VMS username to be used for
Mail access.  Any or all can be a single asterisk wildcard.  The mapping
'*/*/*' would map all authenticated remote usernames directly as equivalent VMS
username.  The mapping 'MARK/*/DANIEL' would map remote user MARK to the VMS
username DANIEL.

A particular username can be disabled by using a hyphen in place of the VMS
username, hence the mapping 'SYSTEM/*/-' disables access to the SYSTEM
account's Mail.

The special-case mapping 'DANIEL/*/(POSTMASTER)' marks the remote username as a
'super-user' of soyMAIL, or Postmaster.


PUBLIC-ACCESS
-------------
Each rule value has the following format

  /path-to-mail/ /<user>/<file>/<folder>/

as in the following example

  /WASD-2006/  /daniel/mail/WASD_2006/

which would map the path '/WASD-2005/' to VMS username 'DANIEL', that accounts
MAIL.MAI mail file, and the folder (case-sensitive) 'WASD_2005'.

Alternatively, to allow mapping to any folder in a public path, with the
folder of the initial access specified

  /path-to-mail/ /<user>/<file>/*<initial-folder>/

as in the following example

  /WASD-2006/  /daniel/mail/*WASD_2006/

with optional restriction of no access to the MAIL and NEWMAIL folders

  /path-to-mail/ /<user>/<file>/*!<initial-folder>/

as in the following example

  /WASD-2006/  /daniel/mail/*!WASD_2006/


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
09-APR-2006  MGD  ConfigPublicAccess() modify wildcard folder to allow
                  initial folder specification and prohibit MAIL and NEWMAIL
02-APR-2006  MGD  ConfigPublicAccess() support wildcard folder
16-MAR-2006  MGD  add [logout-realm]
25-FEB-2006  MGD  add [greeting]
01-FEB-2005  MGD  initial
*/
#endif /* COMMENTS_WITH_COMMENTS */
/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <types.h>
#include <unixlib.h>

/* VMS related header files */
#include <descrip.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <prvdef.h>

/* application header file */
#include "soymail.h"
#include "config.h"

#define FI_LI __FILE__, __LINE__

/* global storage */

int  ConfigTextLength;
char  *ConfigTextPtr;

CONFIG_DATA  SoyMailConfig;

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

/*****************************************************************************/
/*
Load the configuration file into memory as a series of newline-delimitted
strings.  Then leaving that in memory, parse the configuration directives and
set the configuration data structure field.
*/

void ConfigLoad (REQUEST_DATA *rdptr)

{
   BOOL  WasConditional;
   int  status,
        ConditionalCount,
        ConditionalFalseAt;
   char  *cptr, *sptr,
         *NamePtr,
         *ValuePtr,
         *ErrorPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigLoad()\n");

   if (WatchEnabled)
   {
      cptr = TrnLnm("SOYMAIL_CONFIG", NULL, 0);
      WatchThis ("CONFIG FILE !AZ", cptr ? cptr : "logical not defined");
   }
   status = ReadFileIntoMemory ("SOYMAIL_CONFIG", &ConfigTextPtr,
                                                  &ConfigTextLength);

   if (VMSnok (status))
   {
      CgiLibResponseError (FI_LI, status, "SOYMAIL_CONFIG.");
      exit (SS$_NORMAL);
   }

   /* full search is on by default, restrict as required */
   SoyMailConfig.SearchEnabled = TRUE;
   SoyMailConfig.SearchKeywordOnly = FALSE;
   SoyMailConfig.SearchHeaderOnly = FALSE;
   SoyMailConfig.SearchPriority = -1;

   ConditionalCount = ConditionalFalseAt = 0;

   while (*ConfigTextPtr)
   {
      cptr = ConfigTextParse (&ConfigTextPtr, &NamePtr, &ValuePtr);
      if (!cptr) break;

      /****************/
      /* conditionals */
      /****************/

      ErrorPtr = NULL;
      WasConditional = FALSE;

      if (strsame (NamePtr, "if-CGI-", 7))
      {
         WasConditional = TRUE;
         ConditionalCount++;
         CGIVAR (cptr, NamePtr+7);
         if (!cptr) cptr = "";
         if (!ConditionalFalseAt)
            if (!SearchKeyword (cptr, ValuePtr, NULL))
               ConditionalFalseAt = ConditionalCount;
      }
      else
      if (strsame (NamePtr, "if-not-CGI-", 11))
      {
         WasConditional = TRUE;
         ConditionalCount++;
         CGIVAR (cptr, NamePtr+11);
         if (!cptr) cptr = "";
         if (!ConditionalFalseAt)
            if (SearchKeyword (cptr, ValuePtr, NULL))
               ConditionalFalseAt = ConditionalCount;
      }
      else
      if (strsame (NamePtr, "if-private", -1))
      {
         WasConditional = TRUE;
         ConditionalCount++;
         if (!ConditionalFalseAt)
            if (!rdptr->PrivateAccess)
               ConditionalFalseAt = ConditionalCount;
      }
      else
      if (strsame (NamePtr, "if-public", -1))
      {
         WasConditional = TRUE;
         ConditionalCount++;
         if (!ConditionalFalseAt)
            if (!rdptr->PublicAccess)
               ConditionalFalseAt = ConditionalCount;
      }
      else
      if (strsame (NamePtr, "if-END", -1) ||
          strsame (NamePtr, "end-if", -1) ||
          strsame (NamePtr, "endif", -1))
      {
         WasConditional = TRUE;
         if (ConditionalFalseAt == ConditionalCount) ConditionalFalseAt = 0;
         if (ConditionalCount > 0)
            ConditionalCount--;
         else
            ErrorPtr = "conditonal underflow";
      }

      if (WatchEnabled)
      {
         if (WasConditional)
            if (!ConditionalFalseAt || ConditionalCount == ConditionalFalseAt)
               cptr = "TEST";
            else
               cptr = "SKIP";
         else
         if (ConditionalFalseAt)
            cptr = "SKIP";
         else
            cptr = "DO";

         if (!*ValuePtr)
            sptr = "";
         else
         if (strchr (ValuePtr, '\n'))
            sptr = "\n";
         else
            sptr = " ";

         WatchThis ("CONFIG !UL !AZ [!AZ]!AZ!AZ",
                    ConditionalCount, cptr, NamePtr,
                    sptr, ValuePtr);
      }

      if (ConditionalFalseAt || WasConditional)
      {
         if (ErrorPtr) ErrorExit (SS$_BUGCHECK, FI_LI);;
         continue;
      }

      if (strsame (NamePtr, "end", -1)) break;

      /*****************/
      /* set directive */
      /*****************/

      if (strsame (NamePtr, "downtime", -1))
         SoyMailConfig.DowntimeMessage = ValuePtr;
      else
      if (strsame (NamePtr, "message-list-footer", -1))
         SoyMailConfig.MessageListFooter = ValuePtr;
      else
      if (strsame (NamePtr, "font-family-local", -1))
         SoyMailConfig.FontFamilyLocal = ValuePtr;
      else
      if (strsame (NamePtr, "greeting", -1))
         SoyMailConfig.InitialGreeting = ValuePtr;
      else
      if (strsame (NamePtr, "help-about-site", -1))
         SoyMailConfig.HelpAboutSite = ValuePtr;
      else
      if (strsame (NamePtr, "html-editor-load", -1))
         SoyMailConfig.HtmlEditorLoad = ValuePtr;
      else
      if (strsame (NamePtr, "logout-realm", -1))
         SoyMailConfig.LogoutRealm = ValuePtr;
      else
      if (strsame (NamePtr, "page-title", -1) ||
          strsame (NamePtr, "page-title1", -1))
         SoyMailConfig.PageTitle1 = ValuePtr;
      else
      if (strsame (NamePtr, "page-title2", -1))
         SoyMailConfig.PageTitle2 = ValuePtr;
      else
      if (strsame (NamePtr, "print-footer", -1))
         SoyMailConfig.PrintFooter = ValuePtr;
      else
      if (strsame (NamePtr, "print-header", -1))
         SoyMailConfig.PrintHeader = ValuePtr;
      else
      if (strsame (NamePtr, "private-access", -1))
         SoyMailConfig.PrivateAccess = ValuePtr;
      else
      if (strsame (NamePtr, "public-access", -1))
         SoyMailConfig.PublicAccess = ValuePtr;
      else
      if (strsame (NamePtr, "public-options-default", -1))
         SoyMailConfig.PublicOptionsDefault = ValuePtr;
      else
      if (strsame (NamePtr, "search-control", -1))
         SoyMailConfig.SearchControl = ValuePtr;
      else
      if (strsame (NamePtr, "SMTP-default-host", -1))
         SoyMailConfig.SmtpDefaultHost = ValuePtr;
      else
      if (strsame (NamePtr, "SMTP-from-host", -1))
         SoyMailConfig.SmtpFromHost = ValuePtr;
      else
      if (strsame (NamePtr, "SMTP-server-host", -1))
         SoyMailConfig.SmtpServerHost = ValuePtr;
      else
      if (strsame (NamePtr, "soyMAIL-at-title", -1))
         SoyMailConfig.SoyMailAtTitle = ValuePtr;
      else
      if (strsame (NamePtr, "user-options-default", -1))
         SoyMailConfig.UserOptionsDefault = ValuePtr;
      else
      if (strsame (NamePtr, "user-options-override", -1))
         SoyMailConfig.UserOptionsOverride = ValuePtr;
      else
      if (WatchEnabled)
         WatchThis ("CONFIG ERROR [!AZ]", NamePtr);
   }

   if (SoyMailConfig.SearchControl)
   {
      for (cptr = SoyMailConfig.SearchControl; *cptr; cptr++)
         *cptr = tolower(*cptr);
      if (strstr (SoyMailConfig.SearchControl, "full"))
      {
         SoyMailConfig.SearchEnabled = TRUE;
         SoyMailConfig.SearchKeywordOnly = FALSE;
         SoyMailConfig.SearchHeaderOnly = FALSE;
      }
      if (strstr (SoyMailConfig.SearchControl, "keyword-only"))
         SoyMailConfig.SearchKeywordOnly = TRUE;
      if (strstr (SoyMailConfig.SearchControl, "header-only"))
         SoyMailConfig.SearchHeaderOnly = TRUE;
      if (strstr (SoyMailConfig.SearchControl, "vms-select"))
         SoyMailConfig.SelectNotSearch = TRUE;
      if (cptr = strstr (SoyMailConfig.SearchControl, "priority="))
      {
         cptr += 9;
         if (isdigit(*cptr)) SoyMailConfig.SearchPriority = atol(cptr);
      }
      if (strstr (SoyMailConfig.SearchControl, "none"))
      {
         SoyMailConfig.SearchEnabled =
            SoyMailConfig.SearchKeywordOnly =
            SoyMailConfig.SearchHeaderOnly =
            SoyMailConfig.SelectNotSearch = FALSE;
      }
   }

   if (SoyMailConfig.LogoutRealm)
   {
      /* realm has been set */
      if (!strcmp (SoyMailConfig.LogoutRealm, "-"))
      {
         /* setting disables logout */
         SoyMailConfig.LogoutDisabled = TRUE;
      }
   }
   else
   {
      /* realm has not been set */
      if (!CgiLibEnvironmentIsWasd())
      {
         /* Apache and OSU must have it set for logout to function */
         SoyMailConfig.LogoutDisabled = TRUE;
      }
   }

   if (WatchEnabled)
      if (ConditionalCount)
         WatchThis ("CONFIG WARNING conditionals not balanced");
}

/*****************************************************************************/
/*
Parses the contents of a file read into memory, dividing it up into
null-terminated strings with names and values.  The general format for these
name-value pairs is '[name] value', with values allowed to comprise one or more
newline separated strings.

NOTE: This function is used to parse a number of different soyMAIL files that
all use this '[name] value' schema.
*/

char* ConfigTextParse
(
char** TextPtrPtr,
char** NamePtrPtr,
char** ValuePtrPtr
)
{
#undef ISLWS
#define ISLWS(ch) (ch == ' ' || ch == '\t')
#define ISNL(ch) (ch == '\n')

   char  *cptr, *sptr, *nwsptr,
         *NamePtr,
         *ValuePtr,
         *ErrorPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ConfigTextParse()\n");

   if (!TextPtrPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (!NamePtrPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (!ValuePtrPtr) ErrorExit (SS$_BUGCHECK, FI_LI);

   cptr = *TextPtrPtr;
   if (!*cptr) return (NULL);

   NamePtr = ValuePtr = NULL;
   *NamePtrPtr = NamePtr;
   *ValuePtrPtr = ValuePtr;

   while (*cptr)
   {
      /* ignore leading white-space */
      while (*cptr && (ISLWS(*cptr) || ISNL(*cptr))) cptr++;
      if (!*cptr) break;
      if (*cptr == '#')
      {
         while (*cptr && !ISNL(*cptr)) cptr++;
         continue;
      }
      sptr = cptr;
      while (*cptr && *cptr != ']' && !ISNL(*cptr))
      {
         if (*cptr == '\\' && *(cptr+1) && *(cptr+1) != '\n') cptr++;
         cptr++;
      }
      if (*sptr != '[' || *cptr != ']')
      {
         if (WatchEnabled)
            WatchThis ("CONFIG PARSE ERROR !#AZ", cptr-sptr, sptr);
         while (*cptr && !ISNL(*cptr)) cptr++;
         continue;
      }
      NamePtr = ++sptr;
      *cptr++ = '\0';
      /* scan over intervening white-space */
      while (*cptr)
      {
         while (*cptr && ISLWS(*cptr)) cptr++;
         if (*cptr && !ISNL(*cptr)) break;
         if (*(cptr+1) == '[' || *(cptr+1) == '#') break;
         cptr++;
      }
      /* also note first non-white-space */
      ValuePtr = nwsptr = sptr = cptr;
      /* scan to next directive */
      while (*cptr)
      {
         while (*cptr && !ISNL(*cptr))
         {
            if (*cptr == '\\' && *(cptr+1) && *(cptr+1) != '\n') cptr++;
            *sptr = *cptr;
            /* note most recent non-white-space */
            if (*cptr && !ISLWS(*sptr) && !ISNL(*sptr)) nwsptr = sptr;
            sptr++;
            cptr++;
         }
         if (!*cptr) break;
         /* must be a newline */
         if (*(cptr+1) == '[' || *(cptr+1) == '#') break;
         *sptr++ = *cptr++;
      }
      /* before possibly terminating over the top of any newline */
      while (ISNL(*cptr)) cptr++;
      if (sptr > ValuePtr)
      {
         /* terminate the value */
         *(nwsptr+1) = '\0';
      }
      else
         ValuePtr = "";
      break;
   }

   *NamePtrPtr = NamePtr;
   *ValuePtrPtr = ValuePtr;
   *TextPtrPtr = cptr;

   if (WatchEnabled && NamePtr && ValuePtr)
   {
      if (!*ValuePtr)
         sptr = "";
      else
      if (strchr (ValuePtr, '\n'))
         sptr = "\n";
      else
         sptr = " ";

      WatchThis ("PARSE [!AZ]!AZ!AZ", NamePtr, sptr, ValuePtr);
   }

   return (NamePtr);
}

/****************************************************************************/
/*
Write the supplied text to the supplied file ensuring that with multi-line text
any subsequent lines have the leading, configuration-reserved characters '['
and '#', escaped.  Even with empty text writes at least a newline character!
*/ 

int ConfigWriteEscaped
(
FILE *fp,
char *TextPtr
)
{
   int  lcnt, retval;
   char  *cptr, *eptr, *sptr;

   /*********/
   /* begin */
   /*********/

   lcnt = retval = 0;
   cptr = sptr = TextPtr;
   for (;;)
   {
      while (*cptr && *cptr != '\n') cptr++;
      lcnt++;
      if (lcnt > 1 && (*sptr == '[' || *sptr == '#'))
         eptr = "\\";
      else
         eptr = "";
      retval += fprintf (fp, "%s%*.*s\n", eptr, cptr-sptr, cptr-sptr, sptr);
      if (*cptr) cptr++;
      if (!*cptr) break;
      sptr = cptr;
   }
   return (retval);
}

/****************************************************************************/
/*
Map the private access path into the respective VMS Mail username.
Returns TRUE is a mapping can be made, FALSE if not.
*/ 

BOOL ConfigPrivateAccess (REQUEST_DATA *rdptr)

{
   char  *cptr, *sptr, *zptr,
         *AuthRealmPtr,
         *MappingPtr;
   char  AuthRealmName [64],
         RemoteUserName [64],
         VmsUserName [64];

   /*********/
   /* begin */
   /*********/

   CGIVARNULL (AuthRealmPtr, "AUTH_REALM");

   cptr = SoyMailConfig.PrivateAccess;
   if (!cptr) cptr = "";

   while (*cptr)
   {
      /* get mapped remote (authenticated) user name */
      while (*cptr && ISLWS(*cptr)) cptr++;
      MappingPtr = cptr;
      zptr = (sptr = RemoteUserName) + sizeof(RemoteUserName)-1;
      while (*cptr && *cptr != '/' && *cptr != '\n' &&
             !ISLWS(*cptr) && sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      /* get mapped authentication realm name (WASDism) */
      while (*cptr && (ISLWS(*cptr) || *cptr == '/')) cptr++;
      zptr = (sptr = AuthRealmName) + sizeof(AuthRealmName)-1;
      while (*cptr && *cptr != '/' && *cptr != '\n' &&
             !ISLWS(*cptr) && sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      /* get mapped VMS user name */
      while (*cptr && (ISLWS(*cptr) || *cptr == '/')) cptr++;
      zptr = (sptr = VmsUserName) + sizeof(VmsUserName)-1;
      while (*cptr && *cptr != '/' && *cptr != '\n' &&
             !ISLWS(*cptr) && sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      if (WatchEnabled)
         WatchThis ("PRIVATE mapping !#AZ", cptr-MappingPtr, MappingPtr);

      /* go to end-of-line */
      while (*cptr && !ISNL(*cptr)) cptr++;
      while (*cptr && ISNL(*cptr)) cptr++;

      if (WatchEnabled)
         WatchThis ("PRIVATE remote !AZ !AZ",
                    RemoteUserName, rdptr->RemoteUser);

      if (RemoteUserName[0] != '*' &&
          !strsame (RemoteUserName, rdptr->RemoteUser, -1))
         continue;

      if (WatchEnabled && AuthRealmPtr)
         WatchThis ("PRIVATE realm !AZ !AZ", AuthRealmName, AuthRealmPtr);

      if (AuthRealmPtr &&
          AuthRealmName[0] != '*' &&
          !strsame (AuthRealmName, AuthRealmPtr, -1))
         continue;

      if (strsame (VmsUserName, "(POSTMASTER)", -1))
      {
         rdptr->PostMaster = TRUE;
         continue;
      }

      zptr = (sptr = rdptr->UserName) + sizeof(rdptr->UserName)-1;
      if (VmsUserName[0] == '*')
         cptr = rdptr->RemoteUser;
      else
         cptr = VmsUserName;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      if (WatchEnabled)
         WatchThis ("PRIVATE username !AZ", rdptr->UserName);
      return (TRUE);
   }

   if (WatchEnabled)
      WatchThis ("PRIVATE failed", rdptr->UserName);
   return (FALSE);
}

/****************************************************************************/
/*
Map the public access path into the respective VMS Mail username, mail file and
mail folder.  Returns TRUE is a path match is made, FALSE if not.
*/ 

BOOL ConfigPublicAccess (REQUEST_DATA *rdptr)

{
   char  *cptr, *sptr, *tptr, *zptr;
   char  PathMatchBuffer [256];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PUBLIC map !AZ", rdptr->PublicAccessString);

   cptr = SoyMailConfig.PublicAccess;
   if (!cptr) cptr = "";

   while (*cptr)
   {
      while (*cptr && ISLWS(*cptr)) cptr++;
      if (*cptr == '/') cptr++;
      zptr = (sptr = PathMatchBuffer) + sizeof(PathMatchBuffer)-1;
      while (*cptr && *cptr != '/' && !ISLWS(*cptr) && !ISNL(*cptr) &&
             sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      if (!strsame (rdptr->PublicAccessString, PathMatchBuffer, -1))
      {
         /* no match, go to start of next rule */
         if (WatchEnabled) WatchThis ("PUBLIC no !AZ", PathMatchBuffer);
         while (*cptr && !ISNL(*cptr)) cptr++;
         while (ISNL(*cptr)) cptr++;
         continue;
      }

      if (*cptr == '/') cptr++;
      while (*cptr && isspace(*cptr)) cptr++;
      if (*cptr == '/') cptr++;

      zptr = (sptr = rdptr->UserName) + sizeof(rdptr->UserName)-1;
      while (*cptr && *cptr != '/' && !ISLWS(*cptr) && !ISNL(*cptr) &&
             sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      rdptr->UserNameLength = strlen(rdptr->UserName);
      if (*cptr == '/') cptr++;
      while (*cptr && ISLWS(*cptr)) cptr++;

      zptr = (sptr = rdptr->MailFileName) + sizeof(rdptr->MailFileName)-1;
      while (*cptr && *cptr != '/' && !ISLWS(*cptr) && !ISNL(*cptr) &&
             sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      rdptr->MailFileNameLength = strlen(rdptr->MailFileName);
      if (*cptr == '/') cptr++;
      while (*cptr && ISLWS(*cptr)) cptr++;

      if (*cptr == '*')
      {
         /* allowed public access to any folder */
         rdptr->PublicWildcard = TRUE;
         cptr++;
         if (*cptr == '!')
         {
            /* except the MAIL and NEWMAIL folders */
            rdptr->PublicNoNewMail = TRUE;
            cptr++;
         }
      }
      else
      {
         /* t'be sure t'be sure */
         rdptr->PublicWildcard = FALSE;
      }

      /* use the folder specified in the directive */
      zptr = (sptr = rdptr->FolderName) + sizeof(rdptr->FolderName)-1;
      while (*cptr && *cptr != '/' && !ISLWS(*cptr) && !ISNL(*cptr) &&
             sptr < zptr)
      {
         if (*cptr == '\\' && *(cptr+1)) cptr++;
         *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      rdptr->FolderNameLength = strlen(rdptr->FolderName);

      if (*cptr == '/') cptr++;
      while (*cptr && ISLWS(*cptr) && !ISNL(*cptr)) cptr++;
      if (!ISNL(*cptr))
      {
         /* trailing public folder title */
         zptr = (sptr = rdptr->PublicAccessTitle) +
                sizeof(rdptr->PublicAccessTitle)-1;
         while (*cptr && !ISNL(*cptr) && sptr < zptr)
         {
            if (*cptr == '\\' && *(cptr+1)) cptr++;
            *sptr++ = *cptr++;
         }
         *sptr = '\0';
      }

      if (rdptr->UserNameLength &&
          rdptr->MailFileNameLength &&
          (rdptr->FolderNameLength || rdptr->PublicWildcard))
      {
         if (WatchEnabled)
            WatchThis ("PUBLIC as \"!AZ\" \"!AZ\" \"!AZ\" \"!AZ\"",
                       rdptr->UserName, rdptr->MailFileName,
                       rdptr->FolderName, rdptr->PublicAccessTitle);
         return (TRUE);
      }
                                  
      /* something wrong, go to start of next rule */
      if (WatchEnabled) WatchThis ("PUBLIC error !AZ", PathMatchBuffer);
      while (*cptr && *cptr != '\n') cptr++;
      while (*cptr == '\n') cptr++;
   }

   rdptr->UserName[rdptr->UserNameLength = 0] = '\0';
   rdptr->MailFileName[rdptr->MailFileNameLength = 0] = '\0';
   rdptr->FolderName[rdptr->FolderNameLength = 0] = '\0';

   return (FALSE);
}

/*****************************************************************************/

