/*****************************************************************************/
/*

                                 contacts.c

The logical name SOYMAIL_CONTACT_LIST can be used to specify a logical list of
lists.  This works as a multi-value logical name with each value being the
logical name for, or actual file specification of, a soyMAIL format contacts
list or a VMS-style (each line in the file contains a single address) mailing
list. 

For example:

  $ DEFINE /SYSTEM SOYMAIL_CONTACT_LIST -
    ALL_USERS_LIST,GROUP1_USERS_LIST,GROUP2_USERS_LIST,GROUP3_USERS_LIST,-
    EXTERNAL_LIST,MAILING_LIST_LIST

Where ALL_USERS_LIST is a VMS-style, so are GROUP1_USERS_LIST and
GROUP2_USERS_LIST.  MAILING_LIST_LIST is a file containing a VMS-style mailing
list of the mailing lists supported by the system.  For the above it might
contain five lines with:

  ALL_USERS_LIST
  GROUP1_USERS_LIST
  GROUP2_USERS_LIST
  GROUP3_USERS_LIST
  EXTERNAL_LIST

These can then be selected to mail to the entire list.


SOYMAIL CONTACT LIST
--------------------
The soyMAIL contact list file is stored in LDAP Data Interchange Format (LDIF)
format (RFC2849).  It is a very simple implementation based on Mozilla 'export
to LDIF', the RFC, and a number of explanations and examples from 'the Web'. 
It is very-much 'monkey-see, monkey-do'.  But never mind.  It should facilitate
a certain amount of exchange with other agents.  In particular it is designed
to facilitate exchange with Mozilla/Firefox.

Attributes supported by the soyMAIL contact list:

  c                          country
  cn                         common name  (e.g. Mark Daniel)
  description                free text  (multi-line)
  dn                         distinguished name
  givenName                  first name  (e.g. Mark)
  facsimileTelephoneNumber   fax telephone number
  l                          location (city)
  mail                       email address  (primary)
  mobile                     mobile telephone number
  postalCode                 post/ZIP code
  sn                         family name  (e.g. Daniel)
  soyMAILlist                additional email addresses  (multi-line)
  telephoneNumber            telephone number
  pager                      pager telephone number
  postalAddress              street address

soyMAIL consolidates some of these discrete fields into composites (e.g. the cn
and description fields as the 'entry', the postalAddress, l, st, postalCode and
c fields into the 'postal address').


NOTES ON SOYMAIL LDIF FILES
---------------------------
For the SOYMAIL_CONTACTS.LDIF file and [export] generated LDIF content
extraneous empty lines should be avoided.  At least Mozilla 1.7.11 interprets
extraneous empty lines, for instance between comments and after comments before
attributes, as an empty record and inserts such into the address book.  I can't
see that RFC2849 prohibits this sort of usage (except of course where it would
terminate an entry).  So, ensure there are no empty lines except where they
terminate an entry!


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
19-MAR-2006  MGD  refine contact management wrt duplicate common names
15-MAR-2006  MGD  bugfix; ContactsSave() to empty LDIF file
15-FEB-2005  MGD  bugfix; ContactsEdit() no sanity check on un-specified
                    contact, just memset() zero it and continue as 'new' entry
01-FEB-2005  MGD  initial
*/
/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <descrip.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "address.h"
#include "request.h"
#include "config.h"
#include "contacts.h"

#define FI_LI __FILE__, __LINE__

/* global storage */

int  ContactsEntryCount;

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  *CgiEnvironmentPtr;

extern char  CurrentVmsTimeString[],
             SoftwareEnv[],
             SoftwareId[],
             SoftwareVn[];

extern VMS_MAIL_USER VmsMailUser;

/*****************************************************************************/
/*
*/

BOOL ContactsPageRequest (REQUEST_DATA *rdptr)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsPageRequest()\n");

   /* a bit of sanity checking */
   if (!rdptr->PrivateAccess) ErrorExit (SS$_BUGCHECK, FI_LI);

   CGIVARNULL (cptr, "FORM_CONTACT_BTN");
   if (cptr)
   {
      if (rdptr->PrevPageIdent != PAGE_CONTACTS) ErrorExit (SS$_BUGCHECK, FI_LI);

      if (LangSame ("contact_delete", cptr))
      {
         ContactsSave (rdptr);
         ContactsPage (rdptr, FALSE);
      }
      else
      if (LangSame ("contact_edit", cptr))
         ContactsPage (rdptr, FALSE);
      else
      if (LangSame ("contact_export", cptr))
      {
         if (ContactsExport (rdptr)) return (TRUE);
         ContactsPage (rdptr, FALSE);
      }
      else
      if (LangSame ("contact_import", cptr))
      {
         CGIVARNULL (cptr, "FORM_CONTACT_IMPORT_MERGE_FILE");
         if (cptr) ContactsImportLdif (rdptr, cptr);
         ContactsPage (rdptr, TRUE);
      }
      else
      if (LangSame ("contact_merge", cptr))
      {
         CGIVARNULL (cptr, "FORM_CONTACT_IMPORT_MERGE_FILE");
         if (cptr) ContactsMergeLdif (rdptr, cptr);
         ContactsPage (rdptr, TRUE);
      }
      else
      if (LangSame ("print", cptr))
         ContactsPrintView (rdptr);
      else
      if (*cptr == '^')
         ContactsPrintView (rdptr);
      else
      if (LangSame ("contact_save", cptr))
      {
         ContactsSave (rdptr);
         ContactsPage (rdptr, FALSE);
      }
      else
         ErrorExit (SS$_BUGCHECK, FI_LI);
   }
   else
      ContactsPage (rdptr, FALSE);

   return (TRUE);
}

/*****************************************************************************/
/*
Generate the contacts maintenance (edit) page.
*/

void ContactsPage
(
REQUEST_DATA *rdptr,
BOOL ImportMerge
)
{
   int  status;
   REQUEST_DATA  *sdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsPage()\n");

   if (ContactsLoad (rdptr, NULL))
      StatusMessage (FI_LI, 0, LangFor("contact_loaded"),
                     rdptr->ContactsLoadCount,
                     rdptr->ContactsErrorCount,
                     rdptr->ContactsUnknownCount);

   rdptr->FieldFocusPtr = "soymail_form.contact_cn_and_description.focus();";

   MainMenuPageBegin (rdptr, PAGE_CONTACTS);
   MainMenuBar (rdptr);
   StatusInfoPanel (rdptr);

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" \
CELLSPACING=\"0\" ALIGN=\"center\">\n\
<TR><TD STYLE=\"vertical-align:top;\">\n\n");

   ContactsEdit (rdptr, ImportMerge);

   fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n");

   MainMenuPageEnd (rdptr);
}

/*****************************************************************************/
/*
Generate the table containing the contacts including the edit fields for the
selected contact (if any).
*/

void ContactsEdit
(
REQUEST_DATA *rdptr,
BOOL ImportMerge
)
{
   BOOL  EditDone = FALSE;
   int  status,
        ContactCount,
        EditNumber;
   char  *cptr, *sptr, *zptr,
         *LDIFcnPtr,
         *LDIFdescriptionPtr,
         *LDIFfacsimileTelephoneNumberPtr,
         *LDIFmailPtr,
         *LDIFmobilePtr,
         *LDIFtelephoneNumberPtr;
   char  PostalAddress [256];
   CONTACTS_ENTRY  EmptyEntry;
   CONTACTS_ENTRY  *ceptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsEdit()\n");

   muptr = &VmsMailUser;

   fprintf (stdout,
"<!-- contacts begin -->\n\
<TABLE CELLPADDING=\"3\" CELLSPACING=\"1\" ALIGN=\"center\" \
CLASS=\"css_panel_4\">\n\
<TR CLASS=\"css_panel_2\">\
<TD COLSPAN=\"5\" STYLE=\"padding:5px;\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"0\" \
BORDER=\"0\" ALIGN=\"center\" WIDTH=\"100%%\">\n\
<TR CLASS=\"css_panel_2\" \
STYLE=\"font-weight:bold;text-align:center;\">\
<TD WIDTH=\"30%%\">&nbsp;&nbsp;%s</TD>\
<TD WIDTH=\"40%%\">&nbsp;&nbsp;%s</TD>\
<TD WIDTH=\"30%%\">%s&nbsp;&nbsp;</TD>\
</TR>\n\
</TABLE>\n\
</TD></TR>\n",
            LangFor("contact_entry"), 
            LangFor("contact_email"), 
            LangFor("contact_postal")); 

   EditNumber = 0;
   CGIVARNULL (cptr, "FORM_CONTACT_BTN");
   if (LangSame ("contact_edit", cptr))
   {
      CGIVARNULL (cptr, "FORM_CONTACT_EDIT_NUMBER");
      if (cptr) EditNumber = atoi(cptr);
      for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = ceptr->NextEntry)
         if (ceptr->EntryNumber == EditNumber) break;
      /* if select the empty (new) contact */
      if (!ceptr) memset (ceptr = &EmptyEntry, 0, sizeof(CONTACTS_ENTRY));
   }
   else
      memset (ceptr = &EmptyEntry, 0, sizeof(CONTACTS_ENTRY));

   ContactCount = 0;
   while (ceptr)
   {
      if (!(LDIFcnPtr = ceptr->LDIFcnPtr)) LDIFcnPtr = "";
      if (!(LDIFdescriptionPtr = ceptr->LDIFdescriptionPtr))
         LDIFdescriptionPtr = "";
      if (!(LDIFfacsimileTelephoneNumberPtr =
            ceptr->LDIFfacsimileTelephoneNumberPtr))
         LDIFfacsimileTelephoneNumberPtr = "";
      if (!(LDIFmailPtr = ceptr->LDIFmailPtr)) LDIFmailPtr = "";
      if (!(LDIFmobilePtr = ceptr->LDIFmobilePtr)) LDIFmobilePtr = "";
      if (!(LDIFtelephoneNumberPtr = ceptr->LDIFtelephoneNumberPtr))
         LDIFtelephoneNumberPtr = "";

      ContactsPostalAddress (ceptr, PostalAddress, sizeof(PostalAddress));

      if (!EditDone)
      {
         fprintf (stdout,
"<TR CLASS=\"css_panel_4\" VALIGN=\"top\" \
STYLE=\"vertical-align:top;\">\n\
<TD><INPUT TYPE=\"radio\" NAME=\"contact_edit_number\" \
VALUE=\"%d\" CHECKED></TD>\n\
<TD><TEXTAREA CLASS=\"css_text_1\" \
NAME=\"contact_cn_and_description\" ROWS=\"4\" COLS=\"30\" \
WRAP=\"off\" TABINDEX=\"1\">%s%s%s</TEXTAREA>&nbsp;</TD>\n\
<TD><TEXTAREA CLASS=\"css_text_1\" \
NAME=\"contact_mail\" ROWS=\"4\" COLS=\"50\" \
WRAP=\"off\" TABINDEX=\"2\">%s</TEXTAREA></TD>\n\
<TD STYLE=\"white-space:nowrap;text-align:right;padding-right:0;\">\
%s:<BR>%s:<BR>%s:<BR>%s:<BR>%s:</TD>\n\
<TD><TEXTAREA CLASS=\"css_text_1\" \
NAME=\"contact_postaladdress\" ROWS=\"4\" COLS=\"30\" \
WRAP=\"off\"%s TABINDEX=\"3\">%s</TEXTAREA>&nbsp;</TD>\n\
</TR>\n\
\
<TR CLASS=\"css_panel_4\" STYLE=\"vertical-align:top;\">\
<TD></TD><TD COLSPAN=\"4\">\n\
%s:&nbsp;<INPUT CLASS=\"css_text_1\" TYPE=\"text\" \
NAME=\"contact_telephonenumber\" SIZE=\"20\" VALUE=\"%s\" \
TABINDEX=\"4\">&nbsp;\n\
%s:&nbsp;<INPUT CLASS=\"css_text_1\" TYPE=\"text\" \
NAME=\"contact_mobile\" SIZE=\"20\" VALUE=\"%s\" TABINDEX=\"5\">&nbsp;\n\
%s:&nbsp;<INPUT CLASS=\"css_text_1\" TYPE=\"text\" \
NAME=\"contact_facsimiletelephonenumber\" SIZE=\"20\" VALUE=\"%s\" \
TABINDEX=\"6\">&nbsp;\n\
</TD></TR>\n",
                  /* the plus indicate an new (initially empty) entry */
                  *LDIFcnPtr ? EditNumber : 999999,
                  HTML_ESCAPE(LDIFcnPtr),
                  *LDIFdescriptionPtr ? "&#10;" : "",
                  HTML_ESCAPE(LDIFdescriptionPtr),
                  HTML_ESCAPE(LDIFmailPtr),
                  LangFor("contact_postal_street"),
                  LangFor("contact_postal_city"),
                  LangFor("contact_postal_state"),
                  LangFor("contact_postal_code"),
                  LangFor("contact_postal_country"),
                  PostalAddress[0] ? "" : " ONFOCUS=\"this.select()\"",
                  PostalAddress[0] ? HTML_ESCAPE(PostalAddress) : "",
                  LangFor("contact_phone"),
                  HTML_ESCAPE(LDIFtelephoneNumberPtr),
                  LangFor("contact_mobile"),
                  HTML_ESCAPE(LDIFmobilePtr),
                  LangFor("contact_fax"),
                  HTML_ESCAPE(LDIFfacsimileTelephoneNumberPtr));

         ContactsEditButtons (TRUE, ImportMerge);

         /* 'continue' with  the first 'real' entry */
         ceptr = rdptr->ContactsListPtr;
         EditDone = TRUE;
         continue;
      }

      if (ceptr->EntryNumber == EditNumber)
      {
         /* skip over the one we're editing */
         ceptr = ceptr->NextEntry;
         continue;
      }

      fprintf (stdout,
"<TR CLASS=\"css_panel_4\" STYLE=\"vertical-align:top;white-space:normal;\">\
<TD><INPUT TYPE=\"radio\" NAME=\"contact_edit_number\" VALUE=\"%d\"></TD>\
<TD><PRE><U>%s</U>",
               ceptr->EntryNumber,
               HTML_ESCAPE(LDIFcnPtr));

      if (*LDIFdescriptionPtr)
         fprintf (stdout, "\n%s",
                  HTML_ESCAPE(LDIFdescriptionPtr));

      if (*LDIFtelephoneNumberPtr)
         fprintf (stdout, "\n%s: %s",
                  LangFor("contact_phone"),
                  HTML_ESCAPE(LDIFtelephoneNumberPtr));

      if (*LDIFfacsimileTelephoneNumberPtr)
         fprintf (stdout, "\n%s: %s",
                  LangFor("contact_fax"),
                  HTML_ESCAPE(LDIFfacsimileTelephoneNumberPtr));

      if (*LDIFmobilePtr)
         fprintf (stdout, "\n%s: %s",
                  LangFor("contact_mobile"),
                  HTML_ESCAPE(LDIFmobilePtr));

      fprintf (stdout,
"</PRE></TD>\
<TD COLSPAN=\"2\"><PRE>%s</PRE></TD>\
<TD><PRE>%s</PRE></TD>\
</TR>\n",
               HTML_ESCAPE(LDIFmailPtr),
               HTML_ESCAPE(PostalAddress));

      ContactCount++;
 
      /* on to the next entry */
      ceptr = ceptr->NextEntry;
   }

   if (ContactCount > 10) ContactsEditButtons (FALSE, ImportMerge);

   fprintf (stdout,
"</TABLE>\n\
<!--contacts end -->\n\n");
}

/*****************************************************************************/
/*
The [save], [edit], etc., button bar.
*/

void ContactsEditButtons
(
BOOL UnderEdit,
BOOL ImportMerge
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsEditButtons()\n");

   fprintf (stdout,
"<TR CLASS=\"css_panel_2\">\
<TD STYLE=\"text-align:left;padding:5px;\" COLSPAN=\"5\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\" WIDTH=\"100%%\">\
<TR><TD ALIGN=\"left\">\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\"%s>&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\">&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s ... %s\')\">&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_2\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"contact_btn\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\n\
</TD>\n",
            LangFor("contact_save"),
            UnderEdit ? " TABINDEX=\"7\"" : "",
            LangFor("contact_edit"),
            LangFor("contact_delete"),
            LangFor("contact_delete"), LangFor("sure?"), 
            LangFor("print"));

   if (ImportMerge)
   {
      fprintf (stdout,
"<TD ALIGN=\"right\">\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s LDIF ... %s\')\">&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s LDIF ... %s\')\">&nbsp;\n\
<INPUT CLASS=\"css_text_1\" TYPE=\"file\" \
NAME=\"contact_import_merge_file\" SIZE=\"25\">&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\">&nbsp;\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n",
               LangFor("contact_import"),
               LangFor("contact_import"), LangFor("sure?"),
               LangFor("contact_merge"),
               LangFor("contact_merge"), LangFor("sure?"),
               LangFor("contact_export"));
   }
   else
   {
      fprintf (stdout,
"<TD ALIGN=\"right\">\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\">&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\">&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"contact_btn\" VALUE=\"%s\">&nbsp;\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n",
               LangFor("contact_import"),
               LangFor("contact_merge"),
               LangFor("contact_export"));
   }
}

/*****************************************************************************/
/*
The function name say it all.
*/

void ContactsPrintView (REQUEST_DATA *rdptr)

{
   int  status,
        ContactCount;
   char  *cptr,
         *LDIFcnPtr,
         *LDIFdescriptionPtr,
         *LDIFfacsimileTelephoneNumberPtr,
         *LDIFmailPtr,
         *LDIFmobilePtr,
         *LDIFtelephoneNumberPtr;
   char  PostalAddress [256];
   CONTACTS_ENTRY  *ceptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsPrintView()\n");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   ContactsLoad (rdptr, NULL);

   if (rdptr->CharSetPtr) CgiLibResponseSetCharset (rdptr->CharSetPtr);

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" VALUE=\"%s\">\n\
<META NAME=\"environment\" VALUE=\"%s\">\n",
            SoftwareEnv,
            CgiEnvironmentPtr);

   SoyMailStyleStuff (rdptr, "print");

   /* setTimeout() gives Firefox 1.5 a chance to render the page */
   fprintf (stdout,
"<TITLE>%s</TITLE>\n\
</HEAD>\n\n\
<BODY ONLOAD=\"setTimeout(\'print()\',500);\">\n\
\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"5\" BORDER=\"0\">\n\
<TR><TH COLSPAN=\"4\" STYLE=\"font-size:120%%;\">%s</TH></TR>\n\
<TR>\
<TD STYLE=\"padding-right:1em;\"></TD>\
<TH ALIGN=\"left\" STYLE=\"padding-right:1em;\">%s</TH>\
<TH ALIGN=\"left\" STYLE=\"padding-right:1em;\">%s</TH>\
<TH ALIGN=\"left\">%s</TH>\
</TR>\n\
<TR><TD COLSPAN=\"4\" CLASS=\"css_hr\">\
<HR ALIGN=\"left\" SIZE=\"1\" NOSHADE \
CLASS=\"css_print_hr\" STYLE=\"width:100%%\"></TD></TR>\n",
            SoyMailPageTitle (rdptr),
            LangFor("contacts"),
            LangFor("contact_entry"), 
            LangFor("contact_email"),
            LangFor("contact_postal")); 

   ContactCount = 0;
   for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = ceptr->NextEntry)
   {
      ContactCount++;

      if (!(LDIFcnPtr = ceptr->LDIFcnPtr)) LDIFcnPtr = "";
      if (!(LDIFdescriptionPtr = ceptr->LDIFdescriptionPtr))
         LDIFdescriptionPtr = "";
      if (!(LDIFfacsimileTelephoneNumberPtr =
            ceptr->LDIFfacsimileTelephoneNumberPtr))
         LDIFfacsimileTelephoneNumberPtr = "";
      if (!(LDIFmailPtr = ceptr->LDIFmailPtr)) LDIFmailPtr = "";
      if (!(LDIFmobilePtr = ceptr->LDIFmobilePtr)) LDIFmobilePtr = "";
      if (!(LDIFtelephoneNumberPtr = ceptr->LDIFtelephoneNumberPtr))
         LDIFtelephoneNumberPtr = "";

      ContactsPostalAddress (ceptr, PostalAddress, sizeof(PostalAddress));

      fprintf (stdout,
"<TR>\
<TH ALIGN=\"right\" VALIGN=\"top\" STYLE=\"padding-right:1em;\">%d.</TH>\
<TD ALIGN=\"left\" VALIGN=\"top\" STYLE=\"padding-right:1em;\"><PRE><U>%s</U>",
               ContactCount,
               HTML_ESCAPE(LDIFcnPtr));

      if (*LDIFdescriptionPtr)
         fprintf (stdout, "\n%s",
                  HTML_ESCAPE(LDIFdescriptionPtr));

      if (*LDIFtelephoneNumberPtr)
         fprintf (stdout, "\n%s: %s",
                  LangFor("contact_phone"),
                  HTML_ESCAPE(LDIFtelephoneNumberPtr));

      if (*LDIFfacsimileTelephoneNumberPtr)
         fprintf (stdout, "\n%s: %s",
                  LangFor("contact_fax"),
                  HTML_ESCAPE(LDIFfacsimileTelephoneNumberPtr));

      if (*LDIFmobilePtr)
         fprintf (stdout, "\n%s: %s",
                  LangFor("contact_mobile"),
                  HTML_ESCAPE(LDIFmobilePtr));

      fprintf (stdout,
"</PRE></TD>\
<TD ALIGN=\"left\" VALIGN=\"top\" STYLE=\"padding-right:1em;\">\
<PRE>%s</PRE></TD>\
<TD ALIGN=\"left\" VALIGN=\"top\">\
<PRE>%s</PRE></TD>\
</TR>\n",
               HTML_ESCAPE(LDIFmailPtr),
               HTML_ESCAPE(PostalAddress));
   }

   fprintf (stdout,
"<TR><TD COLSPAN=\"4\" CLASS=\"css_hr\" STYLE=\"padding-top:0.8em;\">\
<HR ALIGN=\"left\" SIZE=\"1\" NOSHADE \
CLASS=\"css_print_hr\" STYLE=\"width:100%%\"></TD></TR>\n\
<TR>\
<TH ALIGN=\"left\" COLSPAN=\"2\"  CLASS=\"css_footer\">%s</TD>\n\
<TH ALIGN=\"right\" COLSPAN=\"2\" CLASS=\"css_footer\">%s</TD>\n\
</TR>\n\
\
</TABLE>\n\
\n\
</BODY>\n\
</HTML>\n",
            SOY_MAIL,
            InetMailRfcVmsDate(CurrentVmsTimeString));
}

/*****************************************************************************/
/*
Save the personal contacts to file.
*/

void ContactsSave (REQUEST_DATA *rdptr)

{
   int  retval, status,
        EditNumber;
   char  *cptr, *fnptr, *sptr,
         *ContactCnPtr;
   char  EmptyString [2] = "",
         StringBuffer [512];
   FILE  *fp;
   CONTACTS_ENTRY  *ceptr, *neptr,
                   *EditedEntryPtr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsSave()\n");

   ContactsLoad (rdptr, NULL);

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   fnptr = CallMailFileIn (muptr, NULL, SOY_CONTACTS_LDIF);
   if (WatchEnabled) WatchThis ("CONTACTS !AZ", fnptr);

   EditNumber = 0;
   CGIVARNULL (cptr, "FORM_CONTACT_EDIT_NUMBER");
   if (cptr) EditNumber = atoi(cptr);

   CGIVARNULL (cptr, "FORM_CONTACT_BTN");
   if (LangSame ("contact_delete", cptr))
   {
      /*******************/
      /* delete an entry */
      /*******************/

      /* a record with a zero entry number does not get written out */
      for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = ceptr->NextEntry)
         if (ceptr->EntryNumber == EditNumber)
            ceptr->EntryNumber = 0;
   }
   else
   {
      /************************/
      /* get any edited entry */
      /************************/

      EditedEntryPtr = ceptr = calloc (1, sizeof(CONTACTS_ENTRY));
      if (!ceptr) ErrorExit (vaxc$errno, FI_LI);

      CGIVARNULL (cptr, "FORM_CONTACT_CN_AND_DESCRIPTION");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      while (*cptr && isspace(*cptr)) cptr++;
      ceptr->LDIFcnPtr = ContactCnPtr = cptr;
      /* look for additional lines containing the entry description */
      for (sptr = cptr; *sptr && *sptr != '\r' && *sptr != '\n'; sptr++);
      if (*sptr == '\r') *sptr++ = '\0';
      if (*sptr)
      {
         *sptr++ = '\0';
         /* any description is line 2 and onwards */
         while (*sptr && isspace(*sptr)) sptr++;
         ceptr->LDIFdescriptionPtr = sptr;
      }

      CGIVARNULL (cptr, "FORM_CONTACT_MAIL");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      while (*cptr && isspace(*cptr)) cptr++;
      ceptr->LDIFmailPtr = cptr;

      CGIVARNULL (cptr, "FORM_CONTACT_TELEPHONENUMBER");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      ceptr->LDIFtelephoneNumberPtr = cptr;

      CGIVARNULL (cptr, "FORM_CONTACT_FACSIMILETELEPHONENUMBER");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      ceptr->LDIFfacsimileTelephoneNumberPtr = cptr;

      CGIVARNULL (cptr, "FORM_CONTACT_MOBILE");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      ceptr->LDIFmobilePtr = cptr;

      CGIVARNULL (cptr, "FORM_CONTACT_POSTALADDRESS");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      /* if it's the unmodified template then it's effectively empty */
      if (*cptr == '_') cptr = EmptyString;
      ceptr->LDIFpostalAddressPtr = cptr;
      ceptr->LDIFlPtr = ceptr->LDIFstPtr =
         ceptr->LDIFpostalCodePtr = ceptr->LDIFcPtr = "";

      /* remove that pesky WIN32/DOS carriage-control */
      sptr = cptr;
      while (*cptr)
      {
         while (*cptr == '\r') cptr++;
         if (!*cptr) break;
         *sptr++ = *cptr++;
      }
      *sptr = '\0';

      /* parse this one multi-line text value into components */
      sptr = ceptr->LDIFpostalAddressPtr;
      while (*sptr && *sptr != '\n') sptr++;
      if (*sptr) *sptr++ = '\0';
      /* location (city) */
      ceptr->LDIFlPtr = sptr;
      while (*sptr && *sptr != '\n') sptr++;
      if (*sptr) *sptr++ = '\0';
      /* state */
      ceptr->LDIFstPtr = sptr;
      while (*sptr && *sptr != '\n') sptr++;
      if (*sptr) *sptr++ = '\0';
      /* post/ZIP code */
      ceptr->LDIFpostalCodePtr = sptr;
      while (*sptr && *sptr != '\n') sptr++;
      if (*sptr) *sptr++ = '\0';
      /* country */
      ceptr->LDIFcPtr = sptr;
      while (*sptr && *sptr != '\n') sptr++;
      *sptr = '\0';

      /* if then entry (common) name is not empty */
      if (*ContactCnPtr)
      {
         /* a record with a zero entry number name does not get written out */
         for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = ceptr->NextEntry)
            if (ceptr->EntryNumber == EditNumber)
               ceptr->EntryNumber = 0;

         /* append the new record to the last entry in the list */
         for (ceptr = rdptr->ContactsListPtr;
              ceptr && ceptr->NextEntry;
              ceptr = ceptr->NextEntry);
         if (ceptr)
            ceptr->NextEntry = EditedEntryPtr;
         else
            rdptr->ContactsListPtr = EditedEntryPtr;

         EditedEntryPtr->EntryNumber = ++ContactsEntryCount;
      }
   }

   /****************/
   /* save to file */
   /****************/

   fp = fopen (fnptr, "w");
   if (!fp)
   {
      status = vaxc$errno;
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
      return;
   }

   status = SS$_NORMAL;

   retval = fprintf (fp,
"# soyMAIL contacts in LDIF as at %s\n# [version] %s\n#\n",
                     CurrentVmsTimeString, SoftwareVn);
   if (retval < 0) status = vaxc$errno;

   for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = ceptr->NextEntry)
   {
      /* a record with a zero entry number does not get written out */
      if (!ceptr->EntryNumber) continue;

      /* look for additional lines containing the additional addresses */
      for (sptr = ceptr->LDIFmailPtr;
           *sptr && *sptr != '\r' && *sptr != '\n';
           sptr++);
      if (*sptr)
      {
         /* additional, newline separated addresses are a soyMAIL-ism */
         *sptr++ = '\0';
         while (*sptr && isspace(*sptr)) sptr++;
         ceptr->LDIFsoyMAILlistPtr = sptr;
      }

      PrintFao (StringBuffer, sizeof(StringBuffer),
                "cn=!AZ,mail=!AZ", ceptr->LDIFcnPtr, ceptr->LDIFmailPtr);
      if (Debug) fprintf (stdout, "|%s|\n", StringBuffer);
      status = ContactsLdifRecord (fp, "dn", StringBuffer);
      if (VMSnok (status)) break;

      retval = fprintf (fp,
"objectclass: top\n\
objectclass: person\n\
objectclass: organizationalPerson\n\
objectclass: inetOrgPerson\n");
      if (retval < 0) status = vaxc$errno;
      if (VMSnok (status)) break;

      status = ContactsLdifRecord (fp, "cn", ceptr->LDIFcnPtr);
      if (VMSnok (status)) break;

      if (ceptr->LDIFdescriptionPtr && *ceptr->LDIFdescriptionPtr)
      {
         status = ContactsLdifRecord (fp, "description",
                                      ceptr->LDIFdescriptionPtr);
         if (VMSnok (status)) break;
      }

      status = ContactsLdifRecord (fp, "mail", ceptr->LDIFmailPtr);
      if (VMSnok (status)) break;

      if (ceptr->LDIFsoyMAILlistPtr && *ceptr->LDIFsoyMAILlistPtr)
      {
         status = ContactsLdifRecord (fp, "soyMAILlist", sptr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFtelephoneNumberPtr && *ceptr->LDIFtelephoneNumberPtr)
      {
         status = ContactsLdifRecord (fp, "telephoneNumber",
                                      ceptr->LDIFtelephoneNumberPtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFfacsimileTelephoneNumberPtr &&
          *ceptr->LDIFfacsimileTelephoneNumberPtr)
      {
         status = ContactsLdifRecord (fp, "facsimileTelephoneNumber",
                                      ceptr->LDIFfacsimileTelephoneNumberPtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFmobilePtr && *ceptr->LDIFmobilePtr)
      {
         status = ContactsLdifRecord (fp, "mobile", ceptr->LDIFmobilePtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFpostalAddressPtr && *ceptr->LDIFpostalAddressPtr)
      {
         status = ContactsLdifRecord (fp, "postalAddress",
                                      ceptr->LDIFpostalAddressPtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFlPtr && *ceptr->LDIFlPtr)
      {
         status = ContactsLdifRecord (fp, "l", ceptr->LDIFlPtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFstPtr && *ceptr->LDIFstPtr)
      {
         status = ContactsLdifRecord (fp, "st", ceptr->LDIFstPtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFpostalCodePtr && *ceptr->LDIFpostalCodePtr)
      {   status = ContactsLdifRecord (fp, "postalCode",
                                      ceptr->LDIFpostalCodePtr);
         if (VMSnok (status)) break;
      }

      if (ceptr->LDIFcPtr && *ceptr->LDIFcPtr)
      {
         status = ContactsLdifRecord (fp, "c", ceptr->LDIFcPtr);
         if (VMSnok (status)) break;
      }

      /* entry terminating empty line */
      retval = fprintf (fp, "\n");
      if (retval < 0) status = vaxc$errno;
      if (VMSnok (status)) break;
   }

   retval = fprintf (fp, "# end\n");
   if (retval < 0) status = vaxc$errno;

   fclose (fp);
   if (VMSok (status))
   {
      /* purge the versions back */
      strcat (fnptr, ";-1");
      while (!remove (fnptr));
      StatusMessage (FI_LI, 0, LangFor("contact_saved"));
   }
   else
   {
      /* remove this latest one */
      strcat (fnptr, ";0");
      remove (fnptr);
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
   }

   /* need to reload after saving, so empty the list */
   for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = neptr)
   {
      neptr = ceptr->NextEntry;
      free (ceptr);
   }
   rdptr->ContactsListPtr = NULL;
   if (rdptr->ContactsTextPtr)
   {
      free (rdptr->ContactsTextPtr);
      rdptr->ContactsTextPtr = NULL;
      rdptr->ContactsTextLength = 0;
   }
}

/*****************************************************************************/
/*
Append an LDIF record to the contacts file.
Called from the message read page to add a contact name and address derived
from the source address.
*/

void ContactsAppend
(
REQUEST_DATA *rdptr,
char *LDIFcnPtr,
char *LDIFmailPtr
)
{
   BOOL  NewFile;
   int  retval, status;
   char  *fnptr;
   char  StringBuffer [512];
   FILE  *fp;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsAppend()\n");

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   fnptr = CallMailFileIn (muptr, NULL, SOY_CONTACTS_LDIF);
   if (WatchEnabled) WatchThis ("CONTACTS !AZ", fnptr);

   fp = fopen (fnptr, "r", "shr=put");
   if (!fp)
      NewFile = TRUE;
   else
   {
      fclose (fp);
      NewFile = FALSE;
   }

   fp = fopen (fnptr, "a");
   if (!fp)
   {
      status = vaxc$errno;
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
      return;
   }

   status = SS$_NORMAL;

   if (NewFile)
   {
      retval = fprintf (fp, "# soyMAIL contacts as at %s\n# [version] %s\n#\n",
                        CurrentVmsTimeString, SoftwareVn);
      if (retval < 0) status = vaxc$errno;
   }

   if (VMSok (status))
   {
      retval = fprintf (fp, "# appended %s\n", CurrentVmsTimeString);
      if (retval < 0) status = vaxc$errno;
   }

   if (VMSok (status))
   {
      PrintFao (StringBuffer, sizeof(StringBuffer),
                "cn=!AZ,mail=!AZ", LDIFcnPtr, LDIFmailPtr);
      if (Debug) fprintf (stdout, "|%s|\n", StringBuffer);
      status = ContactsLdifRecord (fp, "dn", StringBuffer);
   }

   if (VMSok (status))
   {
      retval = fprintf (fp,
"objectclass: top\n\
objectclass: person\n\
objectclass: organizationalPerson\n\
objectclass: inetOrgPerson\n");
      if (retval < 0) status = vaxc$errno;
   }

   if (VMSok (status))
      status = ContactsLdifRecord (fp, "cn", LDIFcnPtr);

   if (VMSok (status))
      status = ContactsLdifRecord (fp, "mail", LDIFmailPtr);

   if (VMSok (status))
   {
      /* entry terminating empty line */
      retval = fprintf (fp, "\n");
      if (retval < 0) status = vaxc$errno;
   }

   fclose (fp);

   if (VMSok (status))
   {
      /* purge the versions back */
      strcat (fnptr, ";-1");
      while (!remove (fnptr));
      StatusMessage (FI_LI, 0, LangFor("contact_added"));
   }
   else
   {
      /* remove this latest one */
      strcat (fnptr, ";0");
      remove (fnptr);
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
   }
}

/*****************************************************************************/
/*
Write an LDIF compliant 'name: value' pair.
*/

int ContactsLdifRecord
(
FILE *fp,
char *NamePtr,
char *ValuePtr
)
{
   int  cnt, retval, status;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsLdifRecord() |%s|\n", NamePtr);

   /* if the first character is acceptable */
   cptr = ValuePtr;
   if (*cptr > 0 && *cptr <= 127 &&
       *cptr != 10 && *cptr != 13 &&
       *cptr != ' ' && *cptr != ':' && *cptr != '<')
   {
      /* have a look through the rest of the string */
      while (*cptr > 0 && *cptr <= 127 &&
             *cptr != 10 && *cptr != 13) cptr++;
      /* snare a trailing space */
      if (*(cptr-1) == ' ') cptr -= 1;
   }

   if (!*cptr)
   {
      /* does not require base64 encoding */
      retval = fprintf (fp, "%s: %s\n", NamePtr, ValuePtr);
      if (retval >= 0)
         return (SS$_NORMAL);
      else
         return (vaxc$errno);
   }

   /* requires base64 encoding */
   while (*cptr) cptr++;
   ValuePtr = MimeEncBase64 (ValuePtr, cptr - ValuePtr, "");
   if (!ValuePtr) ErrorExit (SS$_BUGCHECK, FI_LI);

   retval = fprintf (fp, "%s:: ", NamePtr);
   cptr = ValuePtr;
   while (*cptr && retval >= 0)
   {
      sptr = cptr;
      for (cnt = 0; cnt < 76 && *cptr; cnt++) cptr++;
      if (cnt)
         retval = fprintf (fp, "%s%*.*s\n",
                           sptr > ValuePtr ? " " : "", cnt, cnt, sptr);
   }
   if (retval >= 0)
      status = SS$_NORMAL;
   else
      status = vaxc$errno;

   /* free the base64 allocated memory (just being neat) */
   free (ValuePtr);

   return (status);
}

/*****************************************************************************/
/*
*/

BOOL ContactsLoad
(
REQUEST_DATA *rdptr,
char *ContactList
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsLoad()\n");

   /* don't need to do it again! */
   if (rdptr->ContactsListPtr) return (FALSE);

   ContactsLoadFile (rdptr, ContactList);

   return (TRUE);
}

/*****************************************************************************/
/*
Populate the contacts list.  The file is either the personal soyMAIL contacts
file or that specified by the contact list CGI variable. If personal soyMAIL
contact information this text is parsed into a linked list of data structures
pointing to each entry's components.
*/

int ContactsLoadFile
(
REQUEST_DATA *rdptr,
char *ContactList
)
{
   BOOL  soyMailContacts;
   int  len, status;
   char  *cptr, *fnptr, *sptr, *zptr;
   char  FileName [256];
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsLoadFile()\n");

   muptr = &VmsMailUser;

   /* if a specific file has been requested and it's not the personal list */
   cptr = ContactList;
   if (cptr && *cptr && !strsame(cptr,SOY_CONTACT_PERSONAL,-1))
   {
      soyMailContacts = FALSE;
      zptr = (sptr = FileName) + sizeof(FileName)-1;
      while (*cptr && sptr < zptr)
      {
         if (isalnum(*cptr) || *cptr == '$')
            *sptr++ = toupper(*cptr);
         else
            *sptr++ = '_';
         cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';
      fnptr = FileName;
   }
   else
   {
      soyMailContacts = TRUE;
      fnptr = CallMailFileIn (muptr, NULL, SOY_CONTACTS_LDIF);
   }

   if (WatchEnabled) WatchThis ("CONTACTS !AZ", fnptr);

   status = ReadFileIntoMemory (fnptr, &rdptr->ContactsTextPtr,
                                       &rdptr->ContactsTextLength);
   if (VMSok (status))
   {
      if (soyMailContacts)
         ContactsLoadLdif (rdptr, rdptr->ContactsTextPtr);
   }
   else
   if (fnptr == FileName || status != RMS$_FNF)
      StatusMessage (FI_LI, 1, "%s: %s.",
                     LangFor("contacts"), SysGetMsg(status));


   return (status);
}

/*****************************************************************************/
/*
Replace any current list with the entries specified in the supplied LDIF
directives.
*/

void ContactsImportLdif
(
REQUEST_DATA *rdptr,
char *LdifTextPtr
)
{
   CONTACTS_ENTRY  *ceptr, *neptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsImportLdif()\n");

   ContactsLoadLdif (rdptr, LdifTextPtr);

   StatusMessage (FI_LI, 0, LangFor("contact_imported"),
                  rdptr->ContactsLoadCount,
                  rdptr->ContactsErrorCount,
                  rdptr->ContactsUnknownCount);
}

/*****************************************************************************/
/*
Append the entries specified in the supplied LDIF directives to any exisiting
entries in the list.
*/

void ContactsMergeLdif
(
REQUEST_DATA *rdptr,
char *LdifTextPtr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsMergeLdif()\n");

   ContactsLoad (rdptr, NULL);

   ContactsLoadLdif (rdptr, LdifTextPtr);

   StatusMessage (FI_LI, 0, LangFor("contact_merged"),
                  rdptr->ContactsLoadCount,
                  rdptr->ContactsErrorCount,
                  rdptr->ContactsUnknownCount);
}

/*****************************************************************************/
/*
Load the contacts list using LDIF directives contained in the supplied text.
*/

void ContactsLoadLdif
(
REQUEST_DATA *rdptr,
char *LdifTextPtr
)
{
   int  len;
   char  *cptr, 
         *NamePtr,
         *ValuePtr;
   CONTACTS_ENTRY  *ceptr1,
                   *ceptr2;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsLoadLdif() |%s|\n", LdifTextPtr);

   rdptr->ContactsErrorCount = 
      rdptr->ContactsLoadCount =
      rdptr->ContactsUnknownCount = 0;

   /* if not NULL then we are appending to this list */
   for (ceptr1 = rdptr->ContactsListPtr;
        ceptr1 && ceptr1->NextEntry;
        ceptr1 = ceptr1->NextEntry);

   /* allocate the first entry */
   ceptr2 = calloc (1, sizeof(CONTACTS_ENTRY));
   if (!ceptr2) ErrorExit (vaxc$errno, FI_LI);

   while (LdifTextPtr && *LdifTextPtr)
   {
      cptr = ContactsLdifParse (&LdifTextPtr, &NamePtr, &ValuePtr);
      if (!cptr) break;

      if (strsame (NamePtr, "c", -1))   /* country */
         ceptr2->LDIFcPtr = ValuePtr;
      else
      if (strsame (NamePtr, "cn", -1))   /* common name */
         ceptr2->LDIFcnPtr = ValuePtr;
      else
      if (strsame (NamePtr, "description", -1))   /* free text */
         ceptr2->LDIFdescriptionPtr = ValuePtr;
      else
      if (strsame (NamePtr, "dn", -1))   /* distinguished name */
         ceptr2->LDIFdnPtr = ValuePtr;
      else
      if (strsame (NamePtr, "facsimileTelephoneNumber", -1))   /* fax */
         ceptr2->LDIFfacsimileTelephoneNumberPtr = ValuePtr;
      else
      if (strsame (NamePtr, "givenName", -1))   /* given name */
         ceptr2->LDIFgivenNamePtr = ValuePtr;
      else
      if (strsame (NamePtr, "l", -1))   /* location (city) */
         ceptr2->LDIFlPtr = ValuePtr;
      else
      if (strsame (NamePtr, "mail", -1))   /* email address */
         ceptr2->LDIFmailPtr = ValuePtr;
      else
      if (strsame (NamePtr, "mobile", -1))   /* mobile telephone */
         ceptr2->LDIFmobilePtr = ValuePtr;
      else
      if (strsame (NamePtr, "telephoneNumber", -1))   /* fixed telephone */
         ceptr2->LDIFtelephoneNumberPtr = ValuePtr;
      else
      if (strsame (NamePtr, "pager", -1))   /* pager number */
         ceptr2->LDIFpagerPtr = ValuePtr;
      else
      if (strsame (NamePtr, "postalAddress", -1))   /* street address */
         ceptr2->LDIFpostalAddressPtr = ValuePtr;
      else
      if (strsame (NamePtr, "postalCode", -1))   /* post/ZIP code */
         ceptr2->LDIFpostalCodePtr = ValuePtr;
      else
      if (strsame (NamePtr, "sn", -1))   /* family name */
         ceptr2->LDIFsnPtr = ValuePtr;
      else
      if (strsame (NamePtr, "st", -1))   /* state address */
         ceptr2->LDIFstPtr = ValuePtr;
      else
      if (strsame (NamePtr, "soyMAILlist", -1))   /* soyMAIL mailing list */
         ceptr2->LDIFsoyMAILlistPtr = ValuePtr;
      else
      if (!strsame (NamePtr, "objectclass", -1) &&
          !strsame (NamePtr, "modifyTimestamp", -1) &&
          !strsame (NamePtr, "objectclass", -1) &&
          !strsame (NamePtr, "workurl", -1) &&
          !strsame (NamePtr, "homeurl", -1) &&
          !strsame (NamePtr, "custom", 6) &&
          !strstr (cptr, "mozilla"))
      {
         /* unknown/non-Mozilla attribute */
         rdptr->ContactsUnknownCount++;
         if (WatchEnabled)
            WatchThis ("LDIF unknown !AZ:!AZ", NamePtr, ValuePtr);
      }

      /* if not the end of the entry */
      if (*LdifTextPtr && *LdifTextPtr != '\n') continue;

      /****************/
      /* end of entry */
      /****************/

      if (!ceptr2->LDIFdnPtr)
      {
         rdptr->ContactsErrorCount++;
         memset (ceptr2, 0, sizeof(CONTACTS_ENTRY));
         if (WatchEnabled) WatchThis ("LDIF error NO dn:");
         continue;
      }
      if (!ceptr2->LDIFcnPtr)
      {
         rdptr->ContactsErrorCount++;
         memset (ceptr2, 0, sizeof(CONTACTS_ENTRY));
         if (WatchEnabled) WatchThis ("LDIF error NO cn:");
         continue;
      }
      if (!ceptr2->LDIFmailPtr)
      {
         rdptr->ContactsErrorCount++;
         memset (ceptr2, 0, sizeof(CONTACTS_ENTRY));
         if (WatchEnabled) WatchThis ("LDIF error NO mail:");
         continue;
      }
      if (WatchEnabled) WatchThis ("LDIF entry");
      rdptr->ContactsLoadCount++;

      if (ceptr2->LDIFsoyMAILlistPtr)
      {
         /* recreate the mailing list by appending it to primary email */
         len = strlen(ceptr2->LDIFmailPtr) +
               strlen(ceptr2->LDIFsoyMAILlistPtr);
         cptr = calloc (1, len+2);
         sprintf (cptr, "%s\n%s",
                  ceptr2->LDIFmailPtr,
                  ceptr2->LDIFsoyMAILlistPtr);
         ceptr2->LDIFmailPtr = cptr;
      }

      /************************/
      /* create a linked list */
      /************************/

      if (ceptr1)
      {
         ceptr1->NextEntry = ceptr2;
         ceptr2->PrevEntry = ceptr1;
      }
      else
         rdptr->ContactsListPtr = ceptr2;
      ceptr1 = ceptr2;

      /* allocate another */
      ceptr2 = calloc (1, sizeof(CONTACTS_ENTRY));
      if (!ceptr2) ErrorExit (SS$_BUGCHECK, FI_LI);
   }

   /* tidy-up */
   free (ceptr2);

   /* ensure the entries are in order */
   ContactsSort (rdptr);
}

/*****************************************************************************/
/*
Parses the newline delimitted lines of LDIF text, one directive at a time,
into null-terminated LDIF strings with names and values. 
*/

char* ContactsLdifParse
(
char** TextPtrPtr,
char** NamePtrPtr,
char** ValuePtrPtr
)
{
#undef ISLWS
#define ISLWS(ch) (ch == ' ' || ch == '\t')
/* when uploaded from WIN32 each line has <CR><LF> carriage control */
#define ISNL(ch) (ch == '\r' || *cptr == '\n')

   BOOL  Base64Encoded;
   char  *cptr, *sptr,
         *NamePtr,
         *ValuePtr,
         *ErrorPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsLdifParse()\n");

   if (!TextPtrPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (!NamePtrPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (!ValuePtrPtr) ErrorExit (SS$_BUGCHECK, FI_LI);

   cptr = *TextPtrPtr;
   if (!*cptr) return (NULL);

   NamePtr = ValuePtr = NULL;
   *NamePtrPtr = NamePtr;
   *ValuePtrPtr = ValuePtr;

   while (*cptr)
   {
      /* ignore leading white-space */
      while (*cptr && (ISLWS(*cptr) || ISNL(*cptr))) cptr++;
      if (!*cptr) break;
      if (*cptr == '#')
      {
         while (*cptr)
         {
            while (*cptr && !ISNL(*cptr)) cptr++;
            while (ISNL(*cptr)) cptr++;
            if (*cptr != ' ') break;
            cptr++;
         }
         continue;
      }
      NamePtr = cptr;
      while (*cptr && *cptr != ':' && !ISNL(*cptr)) cptr++;
      if (*cptr != ':')
      {
         if (WatchEnabled)
            WatchThis ("LDIF PARSE ERROR !#AZ", cptr-NamePtr, NamePtr);
         while (*cptr)
         {
            while (*cptr && !ISNL(*cptr)) cptr++;
            while (ISNL(*cptr)) cptr++;
            if (*cptr != ' ') break;
            cptr++;
         }
         continue;
      }
      /* terminate at the ':' of the name */
      *cptr++ = '\0';
      if (*cptr == ':')
      {
         Base64Encoded = TRUE;
         cptr++;
      }
      else
         Base64Encoded = FALSE;
      /* scan over intervening white-space */
      while (*cptr && ISLWS(*cptr)) cptr++;
      ValuePtr = sptr = cptr;
      while (*cptr)
      {
         while (*cptr && !ISNL(*cptr)) *sptr++ = *cptr++;
         /* when uploaded from WIN32 each line has <CR><LF> carriage control */
         if (*cptr == '\r') cptr++;
         if (*cptr == '\n') cptr++;
         /* ensure we're left pointing at a newline */
         if (*cptr == '\r') cptr++;
         if (*cptr != ' ') break;
         cptr++;
      }
      *sptr = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", ValuePtr);
      if (Base64Encoded) MimeDecBase64 (ValuePtr);
      break;
   }

   *NamePtrPtr = NamePtr;
   *ValuePtrPtr = ValuePtr;
   *TextPtrPtr = cptr;

   if (WatchEnabled && NamePtr && ValuePtr)
      WatchThis ("LDIF !AZ:!AZ", NamePtr, ValuePtr);

   return (NamePtr);
}

/*****************************************************************************/
/*
Using a simple bubble sort place the contact list entries in ascending order.
*/

void ContactsSort (REQUEST_DATA *rdptr)

{
   CONTACTS_ENTRY  ScratchEntry1;
   CONTACTS_ENTRY  *ceptr1, *ceptr2,
                   *NextEntry2;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsSort()\n");

   for (ceptr1 = rdptr->ContactsListPtr; ceptr1; ceptr1 = ceptr1->NextEntry)
   {
      for (ceptr2 = ceptr1->NextEntry; ceptr2; ceptr2 = ceptr2->NextEntry)
      {
         if (strcmp (ceptr1->LDIFdnPtr, ceptr2->LDIFdnPtr) < 0) continue;
         memcpy (&ScratchEntry1, ceptr1, sizeof(CONTACTS_ENTRY));
         NextEntry2 = ceptr2->NextEntry;
         memcpy (ceptr1, ceptr2, sizeof(CONTACTS_ENTRY));
         memcpy (ceptr2, &ScratchEntry1, sizeof(CONTACTS_ENTRY));
         ceptr1->NextEntry = ScratchEntry1.NextEntry;
         ceptr2->NextEntry = NextEntry2;
      }
   }

   ContactsEntryCount = 0;
   for (ceptr1 = rdptr->ContactsListPtr; ceptr1; ceptr1 = ceptr1->NextEntry)
      ceptr1->EntryNumber = ++ContactsEntryCount;
}

/*****************************************************************************/
/*
For the message composition page generate both the contact list and the list
selector if the logical name is defined.
*/

void ContactsCompose
(
REQUEST_DATA *rdptr,
int ListLength
)
{
   int  idx, lcnt;
   char  *cptr, *sptr,
         *CurrentList, *ContactList,
         *VmsNamePtr;
   CONTACTS_ENTRY  *ceptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsCompose()\n");

   /* check if the logical name is defined */
   ContactList = TrnLnm (SOY_CONTACT_LIST, NULL, 0);
   if (ContactList && ListLength) ListLength--;

   fprintf (stdout,
"<SELECT CLASS=\"css_select_1\" NAME=\"contact_addr\" SIZE=\"%d\">\n\
<OPTION VALUE=\"%s\">&nbsp;%s \n",
            ListLength,
            HTML_ESCAPE(rdptr->ComposeData.SelfPtr),
            LangFor("contact_self"));

   if (rdptr->ContactsListPtr)
   {
      /* soyMAIL list */
      for (ceptr = rdptr->ContactsListPtr; ceptr; ceptr = ceptr->NextEntry)
         fprintf (stdout, "<OPTION VALUE=\"%s\">&nbsp;%s \n",
                  HTML_ESCAPE(ceptr->LDIFmailPtr),
                  HTML_ESCAPE(ceptr->LDIFcnPtr));
   }
   else
   {
      /* VMS-style list */
      cptr = rdptr->ContactsTextPtr;
      if (!cptr) cptr = "";
      while (*cptr)
      {
         VmsNamePtr = cptr;
         while (*cptr == ' ') cptr++;
         while (*cptr && *cptr != '\n') cptr++;
         if (*cptr) *cptr++ = '\0';
         if (!*VmsNamePtr || *VmsNamePtr == '!') continue;

         fprintf (stdout, "<OPTION VALUE=\"%s\">&nbsp;%s \n",
                  HTML_ESCAPE(VmsNamePtr), HTML_ESCAPE(VmsNamePtr));
      }
   }

   fprintf (stdout, "</SELECT>\n");

   if (!ContactList)
   {
      /* there is not contact list logical name defined */
      fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"contact_list\" VALUE=\"%s\">\n",
               SOY_CONTACT_PERSONAL);
      return;
   }

   /* generate the selector for the list of lists */

   CGIVAR (CurrentList, "FORM_CONTACT_LIST");

   fprintf (stdout,
"<P CLASS=\"css_p00\">\n\
<SELECT CLASS=\"css_select_1\" NAME=\"contact_list\">\n\
<OPTION VALUE=\"%s\"%s>&nbsp;%s \n",
            SOY_CONTACT_PERSONAL,
            strsame (CurrentList, SOY_CONTACT_PERSONAL, -1) ? " SELECTED" : "",
            LangFor("contact_personal"));

   idx = 1;
   cptr = ContactList;
   while (cptr)
   {
      fprintf (stdout, "<OPTION VALUE=\"%s\"%s>&nbsp;%s \n",
               HTML_ESCAPE(cptr),
               strsame (cptr, CurrentList, -1) ? " SELECTED" : "",
               HTML_ESCAPE(cptr));
      /* get the next definition (if any) */
      cptr = TrnLnm (SOY_CONTACT_LIST, NULL, idx++);
   }

   fprintf (stdout,
"</SELECT>&nbsp;\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n",
            LangFor("open"));
}

/*****************************************************************************/
/*
Merely read the LDIF format contacts file and present it as a plain text page.
*/

int ContactsExport (REQUEST_DATA *rdptr)

{
   int  status,
        ContactsLength;
   char  *cptr, *fnptr,
         *ContactsPtr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsExport()\n");

   muptr = &VmsMailUser;

   fnptr = CallMailFileIn (muptr, NULL, SOY_CONTACTS_LDIF);

   if (WatchEnabled) WatchThis ("CONTACTS !AZ", fnptr);

   status = ReadFileIntoMemory (fnptr, &ContactsPtr, &ContactsLength);
   if (VMSnok (status))
   {
      StatusMessage (FI_LI, 1, "%s: %s.",
                     LangFor("contacts"), SysGetMsg(status));
      return (FALSE);
   }

   CgiLibResponseHeader (200, "text/plain");

   cptr = ContactsPtr;
   if (strsame (cptr, "# soyMAIL", 9))
   {
      while (*cptr && *cptr != '\n') cptr++;
      if (*cptr) cptr++;
      if (strsame (cptr, "# [version]", 11))
      {
         while (*cptr && *cptr != '\n') cptr++;
         if (*cptr) cptr++;
         if (*cptr == '\n') cptr++;
         if (*cptr == '#')
         {
            cptr++;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
         }
      }
   }

   fprintf (stdout,
"# soyMAIL contacts in LDIF exported %s\n# [version] %s\n#\n",
            CurrentVmsTimeString, SoftwareVn);
   fputs (cptr, stdout);

   free (ContactsPtr);

   return (TRUE);
}

/*****************************************************************************/
/*
Try to extract something that can be used as a contact name from the supplied
address.
*/

char* ContactsNameAs (char *AddrPtr)

{
   static char  ContactName [64];

   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsNameAs() |%s|\n", AddrPtr);

   zptr = (sptr = ContactName) + sizeof(ContactName)-1;
   if (AddressIsRfc (AddrPtr))
   {
      /* internet style address */
      for (cptr = AddrPtr; *cptr && *cptr == ' '; cptr++);
      if (*cptr == '\"')
      {
         /* quoted personal name preceded the address */
         cptr++;
         while (*cptr && *cptr != '\"' && sptr < zptr) *sptr++ = *cptr++;
      }
      else
      if (strchr (AddrPtr, '<'))
      {
         /* possible unquoted personal name precedes <..> delimitted address */
         while (*cptr && *cptr != '<' && sptr < zptr) *sptr++ = *cptr++;
         *sptr-- = '\0';
         while (sptr > ContactName && *sptr == ' ') sptr--;
         if (*sptr != ' ') sptr++; 
      }
   }
   else
   {
      /* native VMS Mail, look for what might be a personal name */
      for (cptr = AddrPtr; *cptr && *cptr != '\"'; cptr++);
      if (*cptr == '\"')
      {
         /* found a personal name */
         cptr++;
         while (*cptr && *cptr != '\"' && sptr < zptr) *sptr++ = *cptr++;
      }
      else
      {
         /* just use the username */
         if (cptr = strstr (AddrPtr, "::")) cptr += 2;
         if (!cptr) cptr = AddrPtr;
         while (*cptr && *cptr != ' ' && sptr < zptr) *sptr++ = *cptr++;
      }
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", ContactName);
   return (ContactName);
}

/*****************************************************************************/
/*
Build the composite postal address from the discrete LDIF attributes.
*/

int ContactsPostalAddress
(
CONTACTS_ENTRY  *ceptr,
char *AddressBuffer,
int SizeofAddressBuffer
)
{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ContactsPostalAddress()\n");

   zptr = (sptr = AddressBuffer) + SizeofAddressBuffer - 1;
   if (!(cptr = ceptr->LDIFpostalAddressPtr)) cptr = "";
   while (*cptr && isspace(*cptr)) cptr++;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (!(cptr = ceptr->LDIFlPtr)) cptr = "";
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr && sptr < zptr) *sptr++ = '\n';
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (!(cptr = ceptr->LDIFstPtr)) cptr = "";
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr && sptr < zptr) *sptr++ = '\n';
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (!(cptr = ceptr->LDIFpostalCodePtr)) cptr = "";
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr && sptr < zptr) *sptr++ = '\n';
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (!(cptr = ceptr->LDIFcPtr)) cptr = "";
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr && sptr < zptr) *sptr++ = '\n';
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   return (sptr - AddressBuffer);
}

/*****************************************************************************/

